/*****************************************************************************
*
* Copyright (C) 1996-1998 Atmel Corporation
*
* File          : main.c
* Compiler      : OLD: IAR EWAAVR 2.26c
*               : NOW: WinAVR with AVRGCC3.3
* Output size   : 
* Created       : 16-jun-99
* Modified      : 5-jan-2000
* Modified      : 17-April-2002 by AS
* Modified      : 1-March-2002 by RAA
* Modified		: 6-Jun-2003 by LL
*
* Support mail  : avr@atmel.com
*
* Description   : This Program allows an AVR with bootloader capabilities to 
*                 Read/write its own Flash/EEprom. To enter Programming mode   
*                 PD4 is checked. If this pin is pulled low, programming mode  
*                 is entered. If not, normal execution is done from $0000 
*                 "reset" vector in Application area.
*                 The PD4 pin should be pulled HIGH by an external Pull-up 
*                 resistor.
*
* Other info    : The defines.h file must be set up for individual devices
*                 using the excel-sheet preprocessor
*
*                 In general the linker file should always be verifyed to match
*                 the used part's boot address and size. Note that memory size
*                 is specified in bytes in the linker file.
****************************************************************************/
#include <avr/io.h>
#include <avr/pgmspace.h>

#include "defines.h"
#include "serial.h"

#include "assembly.h"

// int main(void) __attribute__ ((section (".bootloader")));

int main(void)
{
    unsigned int intval,address1,data;
    unsigned char val, ldata;

    void (*funcptr)( void ) = 0x0000;       // Set up function pointer to RESET vector
    initbootuart();                           // initialize uart
    PROGPORT |= PROGCTRL;                   // enable pull-up on PROGCTRL line on PROGPORT
	address1 = 0;
    
	/*
        Override PROGPIN-Check if PD1 was set to 1 before.
		Note that PD1 is normaly set to 0 on reset.
		With this "feature" you can enter the programmingmode from application
		using something like this:
		
		void (*funcptr)( void ) = 0x1c00;
		PORTD |= _BV(PD1);
        funcptr(); 
    */
    if( PROGMODE || (PORTD & _BV(PD1)))        // If PROGPIN is pulled low: programmingmode. 
    {  
      for(;;)                           
      {
        val=recchar();

        if(val=='a')                        //Autoincrement?
        {
          sendchar('Y');		    //Autoincrement is quicker
        }
 
        else if(val=='A')                   //write address1 
        {
          address1=recchar();                //read address1 8 MSB
          address1=(address1<<8)|recchar();
          
          #ifdef LARGE_MEMORY
          if(address1 >= 0x7FFF)
            RAMPZ = (1<<RAMPZ0);
          else
            RAMPZ = 0;
          #endif            
        
          address1 = address1 << 1;            //convert from word address1 to byte address1
          sendchar('\r');
        }

        else if(val=='c')                   //Write program memory, low byte
        {       
          ldata=recchar();
          sendchar('\r');
        }

        else if(val== 'C')                  //Write program memory, high byte 
        {
          data=ldata | (recchar() << 8);
          
		  fill_temp_buffer(data,(address1)); //call asm routine. 
          
		  address1 += 2;  
          sendchar('\r');
        }
        
		else if(val == 'n')                 //Is multiword write supported? yes
    	{
		  sendchar('Y');
		}

        else if(val== 'M')                  //Write many words at once
        {
	      // Let PGSZ = size of one page of program memory, in words.
	      // 
	      // The 'M' is followed by the length byte, which must be
	      // between 0 and the PGSZ.  The length specifies how many words of
	      // data will follow.  After the length byte, there must an even
	      // number of bytes of data:
	      //   low  byte for address offset 0
	      //   high byte for address offset 0
	      //   low  byte for address offset 1
	      //   high byte for address offset 1, etc.
	      // There can be between 0 and PGSZ.  The exact number is determined
	      // from the length byte.  The response sent back to the programmer is
	      // a copy of the length byte, the address of the first word (2
	      // bytes), then the address of the last word+2 (2 bytes), then a
	      // return.
	      //
	      // NOTE: The 'M' command is equivalent to a series of 'c' and
	      // 'C' commands with autoincrementing address.  You still must stop
	      // at page boundaries and do the page write ('m').
	      unsigned char i, len = recchar();
	      unsigned int startaddr = address1;
		  for (i=0; i<len; i++) {
			ldata = recchar();
			data=ldata|(recchar()<<8);
			fill_temp_buffer(data,address1); //call asm routine. 
			address1+=2;  
		  }
		  sendchar(len);
		  sendchar(startaddr>>8);
		  sendchar(startaddr & 0xff);
		  sendchar(address1>>8);
    	  sendchar(address1 & 0xff);
          sendchar('\r');
        }


        else if(val=='e')                   //Chip erase 
        {
          #ifdef LARGE_MEMORY               //SPM uses Z pointer but the pointer is only 16bit and
          RAMPZ = 0x00;                     //can only address up to 64Kbytes FLASH to adresse higher 
                                            //location it require the use of RAMPZ

          for(address1=0;((address1 < (unsigned int)(APP_END&0xFFFF))|(RAMPZ == 0x00));address1 += PAGESIZE)
          {
           
			write_page(address1,(1<<PGERS) + (1<<SPMEN));              //Perform page erase
            write_page(address1, (1<<REENABLE_RWW_BIT) + (1<<SPMEN));  //Re-enable the RWW section
						
            if(address1 >=(0xFFFF-PAGESIZE)) //Last section on lower 64k segment is erased
               RAMPZ = (1<<RAMPZ0);         //RAMPZ has to incremented into upper 64k segment
          		  }
          RAMPZ = 0x00;                     //Clear RAMPZ pointer
          
          #else

          for(address1=0;address1 < APP_END;address1 += PAGESIZE)  //Application section = 60 pages
          {
		    
            write_page(address1,(1<<PGERS) + (1<<SPMEN));       //Perform page erase
            write_page(address1, (1<<REENABLE_RWW_BIT) + (1<<SPMEN));  //Re-enable the RWW section
            
		  }

          #endif          
          sendchar('\r');        
        }

        else if(val=='l')                   // write lockbits 
        {
          
		  write_lock_bits(recchar());
		  
          sendchar('\r');
        }
       
        else if(val== 'm')                  // write page
        {
		  
          write_page(address1, (1<<PGWRT) + (1<<SPMEN));       //Perform page write
          write_page(address1, (1<<REENABLE_RWW_BIT) + (1<<SPMEN));  //Re-enable the RWW section
		  
          sendchar('\r');
        }
        
        else if((val=='P')||(val=='L'))     // Enter programming mode 
        {
          sendchar('\r');
        }

        else if (val=='p')
        {
          sendchar('S');
        }        
        
        else if(val=='R')                   //Read program memory 
        {        
          
		  intval = read_program_memory(address1,0x00);

          sendchar((char)(intval>>8));      //send MSB  
          sendchar((char)intval);           //send LSB

          address1+=2;                       //SPM uses Z pointer but the pointer is only 16bit
          #ifdef LARGE_MEMORY               //and can only address1 up to 64Kbytes FLASH to adresse
          if(address1 == 0x0000)             //higher location it require the use of RAMPZ
          {
            RAMPZ++;                       //RAMPZ has to incremented into upper 64k of FLASH memory
            RAMPZ &=(1<<RAMPZ0);           //Mask away any non used bits
          }
          #endif

        }
        
        else if (val == 'D')
        {
          EEARL = address1;
          EEARH = (address1 >> 8);
          address1++;
          EEDR = recchar();
          EECR |= (1<<EEMWE);
          EECR |= (1<<EEWE);
          while (EECR & (1<<EEWE))
            ;
          sendchar('\r');
        }

        else if (val == 'd')
        {
          EEARL = address1;
          EEARH = (address1 >> 8);
          address1++;
          EECR |= (1<<EERE);
          sendchar(EEDR);
        }
               
        else if(val=='F')                   // read fuse bits
        {
          #ifdef LARGE_MEMORY
          RAMPZ = 0;
          #endif

          sendchar(read_program_memory(0x0000,0x09));

        }        

        else if(val=='r')                   // read lock bits
        {
          #ifdef LARGE_MEMORY
          RAMPZ = 0;
          #endif

          sendchar(read_program_memory(0x0001,0x09));

        }        

        else if(val=='N')               // read high fuse bits
        {
          #ifdef LARGE_MEMORY
          RAMPZ = 0;
          #endif
 
		  sendchar(read_program_memory(0x0003,0x09));

        }        

        else if(val=='t')                   // Return programmer type 
        {
          sendchar( PARTCODE );
          sendchar( 0 );
        }

        else if ((val=='x')||(val=='y')||(val=='T'))
        {
          recchar();
          sendchar('\r');
        }
       
        else if (val=='S')                  // Return software identifier 
        {
          sendchar('A');
          sendchar('V');
          sendchar('R');
          sendchar('B');
          sendchar('O');
          sendchar('O');
          sendchar('T');
        }                
        
        else if (val=='V')                  // Return Software Version
        {
          sendchar('2');
          sendchar('3');
        }        

        else if (val=='s')                  // Return Signature Byte
        {							
          sendchar( SIGNATURE_BYTE_1 );
          sendchar( SIGNATURE_BYTE_2 );
          sendchar( SIGNATURE_BYTE_3 );
        }    
		
        else if (val=='Z')                      // Start application
        {
          sendchar('\r');
          funcptr();                        // Jump to Reset vector 0x0000 in Application Section
        }
		
        else if(val!=0x1b)                  // if not esc
        {
          sendchar('?');
        }
      }
    }
    else
    {
      funcptr();        					// Jump to Reset vector 0x0000 in Application Section
    }
	return 0;
}
